package h08;

import h08.Exceptions.*;

import java.util.Arrays;

/**
 * Represents a booking management. A booking management oversees booking operations, ensuring validity and handling duplicates.
 */
public class BookingManagement {

    /**
     * The bookings to be managed.
     */
    private Booking[] bookings;

    /**
     * The current number of bookings.
     */
    private int size;

    /**
     * The flight management for the bookings.
     */
    private FlightManagement flightManagement;

    /**
     * Constructs a new booking management with the specified initial capacity and flight management.
     *
     * @param initialCapacity  the initial number of bookings that can be managed
     * @param flightManagement the flight management for the bookings
     */
    public BookingManagement(int initialCapacity, FlightManagement flightManagement) {
        this.bookings = new Booking[initialCapacity];
        this.size = 0;
        this.flightManagement = flightManagement;
    }

    /**
     * Creates a booking.
     *
     * @param bookingId    the booking ID of the booking
     * @param flightNumber the flight number of the booking
     * @param passengerId  the passenger ID of the booking
     */
    public void createBooking(String bookingId, String flightNumber, String passengerId) throws InvalidBookingException {
        //TODO H8.5.5
        try {
            validateAndCheckBooking(bookingId, flightNumber, passengerId);

            var flight = flightManagement.getFlight(flightNumber);
                flight.bookSeat();

            if (bookings.length <= size) {
                bookings = Arrays.copyOf(bookings, size*2);
            }

            bookings[size++] = new Booking(bookingId, flightNumber, passengerId);

        } catch (NoSeatsAvailableException ignored) {
        } catch (DuplicateBookingException e) {
            System.out.println("Booking already exists: " + bookingId);
        } catch (InvalidBookingException e) {
            System.out.println("Invalid booking details: " + e.getMessage());
        }
    }

    /**
     * Validates the booking details and checks for duplicates.
     *
     * @param bookingId    the booking ID of the booking
     * @param flightNumber the flight number of the booking
     * @param passengerId  the passenger ID of the booking
     * @throws InvalidBookingException if the booking details are invalid
     * @throws DuplicateBookingException if the booking ID is already in use
     */
    private void validateAndCheckBooking(String bookingId, String flightNumber, String passengerId) throws InvalidBookingException {
        //TODO H8.5.2
        if (bookingId == null || flightNumber == null || passengerId == null || bookingId.isEmpty() || flightNumber.isEmpty() || passengerId.isEmpty()) {
            throw new InvalidBookingException("Invalid booking details provided.");
        }

       try {
           searchBooking(bookingId);
           throw new DuplicateBookingException("A booking with this ID already exists.");
       } catch (BookingNotFoundException ignored) {
       }
    }

    /**
     * Searches for a booking by booking ID.
     *
     * @param bookingId the booking ID of the booking
     * @return the booking with the specified booking ID
     * @throws BookingNotFoundException if the booking ist not found
     */
    private Booking searchBooking(String bookingId) throws BookingNotFoundException {
        //TODO H8.5.3
        for (int i = 0; i < size; i++) {
            if (bookings[i].getBookingId().equals(bookingId)) {
                return bookings[i];
            }
        }

        throw new BookingNotFoundException("Booking not found: " + bookingId);
    }

    /**
     * Returns a booking by booking ID.
     *
     * @param bookingId the booking ID of the booking
     * @return the booking with the specified booking ID
     */
    public Booking getBooking(String bookingId) {
        //TODO H8.5.3
        try {
            return searchBooking(bookingId);
        } catch (BookingNotFoundException e) {
            System.out.println("Error retrieving booking: " + e.getMessage());
            return null;
        }
    }

    /**
     * Cancels a booking by booking ID.
     *
     * @param bookingId the booking ID of the booking
     */
    public void cancelBooking(String bookingId) {
        //TODO H8.5.4
        try {
            searchBooking(bookingId).cancelBooking();
            System.out.println("Booking cancelled successfully: " + bookingId);
        } catch (BookingAlreadyCancelledException e) {
            System.out.println("Already cancelled booking: " + bookingId);
        } catch (BookingNotFoundException e) {
            System.out.println("Error cancelling booking: " + e.getMessage());
        }
    }
}
